# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Impact model."""

from __future__ import annotations

import contextlib
import logging
from datetime import UTC, datetime
from inspect import signature
from os import cpu_count
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Self, cast
from warnings import warn

import jax.numpy as jnp
import numpy as np
import numpy.typing as npt
import xarray as xr
from jax import (
    Array,
    default_backend,
    device_get,
    device_put,
    jit,
    local_device_count,
    make_mesh,
    random,
)
from jax.sharding import Mesh, NamedSharding, PartitionSpec
from jax.typing import ArrayLike
from numpyro.handlers import do, seed, substitute, trace
from numpyro.infer import MCMC, SVI
from numpyro.infer.svi import SVIRunResult, SVIState
from tqdm.auto import tqdm
from xarray import open_zarr
from zarr import open_group

from aimz.model._core import BaseModel
from aimz.model.kernel_spec import KernelSpec
from aimz.sampling._forward import _sample_forward
from aimz.utils._format import _dict_to_datatree, _make_attrs
from aimz.utils._kwargs import _group_kwargs
from aimz.utils._output import (
    _shutdown_writer_threads,
    _start_writer_threads,
    _writer,
)
from aimz.utils._validation import (
    _check_is_fitted,
    _validate_group,
    _validate_kernel_body,
    _validate_X_y_to_jax,
)
from aimz.utils.data._input_setup import _setup_inputs
from aimz.utils.data._sharding import (
    _create_sharded_log_likelihood,
    _create_sharded_sampler,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping, Sized

    from aimz.utils.data import ArrayLoader

logger = logging.getLogger(__name__)


class ImpactModel(BaseModel):
    """A class for impact modeling."""

    def __init__(
        self,
        kernel: Callable,
        rng_key: ArrayLike,
        inference: SVI | MCMC,
        *,
        param_input: str = "X",
        param_output: str = "y",
    ) -> None:
        """Initialize an ImpactModel instance.

        Args:
            kernel: A probabilistic model with NumPyro primitives.
            rng_key (ArrayLike): A pseudo-random number generator key.
            inference: An inference method supported by NumPyro, such as an instance of
                :external:class:`~numpyro.infer.svi.SVI` or
                :external:class:`~numpyro.infer.mcmc.MCMC`.
            param_input: Name of the parameter in the ``kernel`` for the main input
                data.
            param_output: Name of the parameter in the ``kernel`` for the output data.

        Warning:
            The ``rng_key`` parameter should be provided as a **typed key array**
            created with :external:func:`jax.random.key`, rather than a legacy
            ``uint32`` key created with :external:func:`jax.random.PRNGKey`.
        """
        super().__init__(kernel, param_input, param_output)
        self._kernel_spec: KernelSpec | None = None
        if isinstance(rng_key, Array) and rng_key.dtype == jnp.uint32:
            msg = "Legacy `uint32` PRNGKey detected; converting to a typed key array."
            warn(msg, category=UserWarning, stacklevel=2)
            rng_key = random.wrap_key_data(rng_key)
        self._rng_key = rng_key
        if not isinstance(inference, (SVI, MCMC)):
            msg = (
                f"Unsupported inference object: `{type(inference).__name__}`. "
                "Expected `SVI` or `MCMC` from `numpyro.infer`."
            )
            raise TypeError(msg)
        self.inference = inference
        self._vi_result: SVIRunResult | None = None
        self._vi_state = None
        self._posterior: dict[str, Array] | None = None
        self._init_runtime_attrs()

    def _init_runtime_attrs(self) -> None:
        """Initialize runtime attributes."""
        self._fn_vi_update: Callable | None = None
        self._fn_sample_prior_predictive: Callable | None = None
        self._fn_sample_posterior_predictive: Callable | None = None
        self._fn_log_likelihood: Callable | None = None
        self._mesh: Mesh | None
        self._device: NamedSharding | None
        num_devices = local_device_count()
        if num_devices > 1:
            self._mesh = make_mesh((num_devices,), ("obs",))
            self._device = NamedSharding(self._mesh, PartitionSpec("obs"))
        else:
            self._mesh = None
            self._device = None
        self._temp_dir: TemporaryDirectory | None = None
        logger.info(
            "Backend: %s, Devices: %d",
            default_backend(),
            num_devices,
        )

    def __str__(self) -> str:
        """Return a summary of the ImpactModel instance."""
        out = [
            "<ImpactModel>\n",
            f"Kernel: {getattr(self.kernel, '__name__', type(self.kernel).__name__)}",
            f"Inference method: {self.inference.__class__.__name__}",
            f"Input parameter: '{self.param_input}'",
            f"Output parameter: '{self.param_output}'",
            f"Fitted: {getattr(self, '_is_fitted', False)}",
        ]
        outdir = getattr(self, "temp_dir", None)
        if outdir:
            out.append(f"Output directory: {outdir}")

        return "\n".join(out)

    def __repr__(self) -> str:
        """Return a representation of the ImpactModel instance."""
        out = [
            "<ImpactModel",
            (
                f"kernel_name="
                f"{getattr(self.kernel, '__name__', type(self.kernel).__name__)};"
            ),
            f"rng_key_data={random.key_data(self._rng_key)};",
            f"inference_method={self.inference.__class__.__name__};",
            f"param_input={self.param_input!r};",
            f"param_output={self.param_output!r};",
            f"kernel_spec={self.kernel_spec!r};",
            f"fitted={getattr(self, '_is_fitted', False)};",
            f"device={self._device};",
            f"temp_dir={getattr(self, 'temp_dir', None)!r}>",
        ]

        return " ".join(out)

    def __del__(self) -> None:
        """Clean up the temporary directory when the instance is deleted."""
        with contextlib.suppress(AttributeError):
            self.cleanup()
        # Call the parent's __del__ method only if it exists and is callable
        super_del = getattr(super(), "__del__", None)
        if callable(super_del):
            super_del()

    def __getstate__(self) -> dict:
        """Return the state of the object excluding runtime attributes.

        Returns:
            The state of the object, excluding runtime attributes.
        """
        return {
            k: v
            for k, v in self.__dict__.items()
            if not (
                k.startswith("_fn")
                or k in {"_device", "_mesh", "_num_devices", "_temp_dir"}
            )
        }

    def __setstate__(self, state: dict[str, object]) -> None:
        """Restore the state and reinitialize runtime attributes.

        Args:
            state: The state to restore, excluding the runtime attributes.
        """
        self.__dict__.update(state)
        self._init_runtime_attrs()

    @property
    def kernel(self) -> Callable:
        """A probabilistic model with NumPyro primitives."""
        return self._kernel

    @property
    def kernel_spec(self) -> KernelSpec | None:
        """The cached :class:`~aimz.model.KernelSpec` or ``None`` if not yet built."""
        return self._kernel_spec

    @property
    def rng_key(self) -> ArrayLike:
        """Pseudo-random number generator key."""
        return self._rng_key

    @property
    def param_input(self) -> str:
        """Parameter name in :attr:`~aimz.ImpactModel.kernel` for the input data."""
        return self._param_input

    @property
    def param_output(self) -> str:
        """Parameter name in :attr:`~aimz.ImpactModel.kernel` for the output data."""
        return self._param_output

    @property
    def vi_result(self) -> SVIRunResult | None:
        """Variational inference result, or ``None`` if not set.

        :setter: This sets :external:data:`~numpyro.infer.svi.SVIRunResult` and marks
            the model as fitted. It does not perform posterior sampling — use
            :meth:`~aimz.ImpactModel.sample` separately to obtain samples.
        """
        return self._vi_result

    @vi_result.setter
    def vi_result(self, vi_result: SVIRunResult) -> None:
        """Set the variational inference result manually.

        Args:
            vi_result (SVIRunResult): The result from a prior variational inference run.
                It must be a NamedTuple or similar object with the following fields:
                - params (dict): Learned parameters from inference.
                - state (SVIState): Internal SVI state object.
                - losses (ArrayLike): Loss values recorded during optimization.
        """
        if np.any(np.isnan(vi_result.losses)):
            msg = "Loss contains NaN or Inf, indicating numerical instability."
            warn(msg, category=RuntimeWarning, stacklevel=2)

        self._is_fitted = True

        self._vi_result = vi_result

    @property
    def posterior(self) -> dict[str, Array] | None:
        """Posterior samples by variable name, or ``None`` if not set."""
        return self._posterior

    @property
    def temp_dir(self) -> str | None:
        """Temporary directory path, or ``None`` if not set."""
        return self._temp_dir.name if self._temp_dir else None

    def _build_kernel_spec(
        self,
        args_bound: Mapping[str, object],
        *,
        with_output: bool,
    ) -> None:
        """Trace the kernel (if needed) and cache default return sites.

        This method is idempotent: if a compatible spec already exists it is a no-op.
        A compatible spec means: ``with_output`` is ``False`` and we already traced
        once, or ``with_output`` is ``True`` and the existing spec was built with an
        observed output (``output_observed=True``).

        Args:
            args_bound: Mapping of fully bound keyword arguments to invoke the kernel
                (includes the input and, when ``with_output`` is ``True``, the observed
                output variable).
            with_output: If ``True`` the trace is validated expecting the output site to
                be observed. If ``False`` the trace may omit an observed output (e.g.,
                prior predictive).
        """
        if self._kernel_spec:
            if with_output:
                if self._kernel_spec.output_observed:
                    return
            elif self._kernel_spec.traced:
                return
        model_trace = trace(seed(self.kernel, rng_seed=self.rng_key)).get_trace(
            **args_bound,
        )
        _validate_kernel_body(
            self.kernel,
            param_output=self.param_output,
            model_trace=model_trace,
            with_output=with_output,
        )
        sample_sites = tuple(k for k, v in model_trace.items() if v["type"] == "sample")
        return_sites = (
            self.param_output,
            *tuple(k for k, v in model_trace.items() if v["type"] == "deterministic"),
        )
        output_observed = bool(with_output)
        self._kernel_spec = KernelSpec(
            traced=True,
            sample_sites=sample_sites,
            return_sites=return_sites,
            output_observed=output_observed,
        )

    def _coerce_return_sites(
        self,
        return_sites: str | Iterable[str] | None,
    ) -> tuple[str, ...]:
        """Return a normalized tuple of site names.

        Args:
            return_sites: User-provided site name(s) or ``None``.

        Returns:
            A tuple of site names.
        """
        if return_sites is None:
            return cast("KernelSpec", self._kernel_spec).return_sites
        if isinstance(return_sites, str):
            return (return_sites,)

        return tuple(str(s) for s in return_sites)

    def _create_output_subdir(self, output_dir: str | Path | None) -> tuple[Path, Path]:
        """Create a subdirectory for storing output.

        This function is called for its side effect: it creates a subdirectory within
        the specified output directory with a timestamp.

        Args:
            output_dir: The directory where the output subdirectory will be created.

        Returns:
            The paths to the output directory and the created subdirectory.
        """
        if output_dir is None:
            if self._temp_dir is None:
                self._temp_dir = TemporaryDirectory()
                logger.info("Temporary directory created at: %s", self._temp_dir.name)
            output_dir = self._temp_dir.name
            logger.info(
                "No output directory provided. Using the model's temporary directory "
                "for storing output.",
            )
        output_dir = Path(output_dir).expanduser().resolve()
        output_dir.mkdir(parents=True, exist_ok=True)
        timestamp = datetime.now(UTC).strftime("%Y%m%dT%H%M%S%fZ")
        output_subdir = output_dir / timestamp
        output_subdir.mkdir(parents=False, exist_ok=False)

        return output_dir, output_subdir

    def _sample_and_write(
        self,
        num_samples: int,
        rng_key: ArrayLike,
        return_sites: tuple[str, ...],
        output_dir: Path,
        kernel: Callable,
        sampler: Callable,
        samples: dict[str, Array],
        dataloader: ArrayLoader,
        pbar: tqdm,
        **kwargs: object,
    ) -> None:
        """Draw samples using a predictive function and write them concurrently to disk.

        This function iterates over batches from the provided data loader, calls the
        specified sampling function (``sampler``) to generate predictions conditioned on
        the provided ``samples``, and writes the resulting arrays to a Zarr group.

        Args:
            num_samples: Number of samples to draw.
            rng_key: Pseudo-random number generator key.
            return_sites: Names of variables (sites) to return.
            output_dir: Directory where outputs will be saved.
            kernel: Probabilistic model with NumPyro primitives.
            sampler: Function performing predictive sampling; must accept the same
                signature as this function.
            samples: Arrays to condition predictions on.
            dataloader: Iterator over batches of input data. Each batch must be a tuple
                containing a dictionary of arrays and a padding value.
            pbar: Progress bar instance to display sampling progress.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Raises:
            Exception: Any exception raised during sampling or writing is logged, the
                output directory is cleaned up, and the exception is re-raised.
        """
        kwargs_array, kwargs_extra = _group_kwargs(kwargs)

        rng_key, *subkeys = random.split(rng_key, num=len(dataloader) + 1)
        if self._device and self._mesh:
            subkeys = device_put(
                subkeys,
                NamedSharding(self._mesh, PartitionSpec()),
            )

        zarr_group = open_group(output_dir, mode="w")
        zarr_arr = {}
        threads, queues, error_queue = _start_writer_threads(
            return_sites,
            group_path=output_dir,
            writer=_writer,
            queue_size=min(cpu_count() or 1, 4),
        )
        try:
            for (batch, n_pad), subkey in zip(dataloader, subkeys, strict=True):
                kwargs_batch = [
                    v
                    for k, v in batch.items()
                    if k not in (self.param_input, self.param_output)
                ]
                dict_arr = device_get(
                    sampler(
                        kernel,
                        num_samples,
                        subkey,
                        return_sites,
                        samples,
                        self.param_input,
                        kwargs_array._fields + kwargs_extra._fields,
                        batch[self.param_input],
                        *(*kwargs_batch, *kwargs_extra),
                    ),
                )
                for site, arr in dict_arr.items():
                    if site not in zarr_arr:
                        zarr_arr[site] = zarr_group.create_array(
                            name=site,
                            shape=(num_samples, 0, *arr.shape[2:]),
                            dtype="float32" if arr.dtype == "bfloat16" else arr.dtype,
                            chunks=(
                                num_samples,
                                dataloader.batch_size,
                                *arr.shape[2:],
                            ),
                            dimension_names=(
                                "draw",
                                *tuple(
                                    f"{site}_dim_{i}"
                                    for i in range(max(arr.ndim - 1, 1))
                                ),
                            ),
                        )
                    queues[site].put(
                        (arr[:, None] if arr.ndim == 1 else arr)[:, : -n_pad or None],
                    )

                if not error_queue.empty():
                    _, exc, tb = error_queue.get()
                    raise exc.with_traceback(tb)
                pbar.update()
            pbar.set_description("Sampling complete, writing in progress...")
            _shutdown_writer_threads(threads, queues)
        except:
            _shutdown_writer_threads(threads, queues)
            logger.exception(
                "Exception encountered. Cleaning up output directory: %s",
                output_dir,
            )
            rmtree(output_dir, ignore_errors=True)
            raise
        finally:
            pbar.close()

    def sample_prior_predictive_on_batch(
        self,
        X: ArrayLike,
        *,
        num_samples: int = 1000,
        rng_key: ArrayLike | None = None,
        return_sites: str | Iterable[str] | None = None,
        return_datatree: bool = True,
        **kwargs: object,
    ) -> xr.DataTree | dict[str, npt.NDArray]:
        """Draw samples from the prior predictive distribution.

        Args:
            X (ArrayLike): Input data with shape ``(n_samples_X, n_features)``.
            num_samples: The number of samples to draw.
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed.
            return_sites: Names of variables (sites) to return. If ``None``, samples
                :attr:`~aimz.ImpactModel.param_output` and deterministic sites.
            return_datatree: If ``True``, return a :external:class:`~xarray.DataTree`;
                otherwise return a :class:`dict`.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            Prior predictive samples. Posterior samples are included if available.

        Raises:
            TypeError: If :attr:`~aimz.ImpactModel.param_output` is passed as an
                argument.
        """
        X = cast("Array", _validate_X_y_to_jax(X))

        # Validate the provided parameters against the kernel's signature
        args_bound = (
            signature(self.kernel).bind(**{self.param_input: X, **kwargs}).arguments
        )
        if self.param_output in args_bound:
            msg = f"Specifying {self.param_output!r} is not allowed."
            raise TypeError(msg)
        self._build_kernel_spec(args_bound, with_output=False)

        if rng_key is None:
            self._rng_key, rng_key = random.split(self._rng_key)

        prior_predictive_samples = device_get(
            _sample_forward(
                self.kernel,
                rng_key=rng_key,
                num_samples=num_samples,
                return_sites=self._coerce_return_sites(return_sites),
                samples=None,
                model_kwargs=args_bound,
            ),
        )

        if not return_datatree:
            return prior_predictive_samples

        out = xr.DataTree(name="root")
        out["prior_predictive"] = _dict_to_datatree(prior_predictive_samples)
        if self.posterior:
            out["posterior"] = _dict_to_datatree(self.posterior)

        return out

    def sample_prior_predictive(
        self,
        X: ArrayLike,
        *,
        num_samples: int = 1000,
        rng_key: ArrayLike | None = None,
        return_sites: str | Iterable[str] | None = None,
        batch_size: int | None = None,
        output_dir: str | Path | None = None,
        progress: bool = True,
        **kwargs: object,
    ) -> xr.DataTree:
        """Draw samples from the prior predictive distribution.

        Results are written to disk in the Zarr format, with computing and file writing
        decoupled and executed concurrently.

        Args:
            X (ArrayLike): Input data with shape ``(n_samples_X, n_features)``.
            num_samples: The number of samples to draw.
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed.
            return_sites: Names of variables (sites) to return. If ``None``, samples
                :attr:`~aimz.ImpactModel.param_output` and deterministic sites.
            batch_size: The batch size for data loading during prior predictive
                sampling. It also determines the chunk size used to store the samples.
                If ``None``, it is determined automatically based on the input data and
                number of samples.
            output_dir: The directory where the outputs will be saved. If the specified
                directory does not exist, it will be created automatically. If ``None``,
                a default temporary directory will be created. A timestamped
                subdirectory will be generated within this directory to store the
                outputs. Outputs are saved in the Zarr format.
            progress: Whether to display a progress bar.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            Prior predictive samples. Posterior samples are included if available.

        Raises:
            TypeError: If :attr:`~aimz.ImpactModel.param_output` is passed as an
                argument.

        See Also:
            :meth:`~aimz.ImpactModel.cleanup` to remove the temporary directory if
            created.
        """
        # Validate the provided parameters against the kernel's signature
        args_bound = (
            signature(self.kernel).bind(**{self.param_input: X, **kwargs}).arguments
        )
        if self.param_output in args_bound:
            msg = f"Specifying {self.param_output!r} is not allowed."
            raise TypeError(msg)

        return_sites = self._coerce_return_sites(return_sites)

        # Prior sampling
        dataloader, _ = _setup_inputs(
            X=X,
            y=None,
            rng_key=self.rng_key,
            batch_size=self._device.num_devices if self._device is not None else 1,
            num_samples=num_samples,
            shuffle=False,
            device=self._device,
            **kwargs,
        )
        batch, _ = next(iter(dataloader))
        self._build_kernel_spec(batch, with_output=False)
        if rng_key is None:
            self._rng_key, rng_key = random.split(self._rng_key)
        rng_key, rng_subkey = random.split(rng_key)
        prior_samples = _sample_forward(
            self.kernel,
            rng_key=rng_subkey,
            num_samples=num_samples,
            return_sites=None,
            samples=None,
            model_kwargs=batch,
        )
        prior_samples = {
            k: v for k, v in prior_samples.items() if k not in return_sites
        }

        kwargs_array, kwargs_extra = _group_kwargs(kwargs)
        if self._fn_sample_prior_predictive is None:
            self._fn_sample_prior_predictive = _create_sharded_sampler(
                self._mesh,
                len(kwargs_array),
                len(kwargs_extra),
            )

        output_dir, output_subdir = self._create_output_subdir(output_dir)

        dataloader, _ = _setup_inputs(
            X=X,
            y=None,
            rng_key=self.rng_key,
            batch_size=batch_size,
            num_samples=num_samples,
            shuffle=False,
            device=self._device,
            **kwargs,
        )

        self._sample_and_write(
            num_samples,
            rng_key=rng_key,
            return_sites=return_sites,
            output_dir=output_subdir,
            kernel=self.kernel,
            sampler=self._fn_sample_prior_predictive,
            samples=prior_samples,
            dataloader=dataloader,
            pbar=tqdm(
                desc=(f"Prior predictive sampling [{', '.join(return_sites)}]"),
                total=len(dataloader),
                disable=not progress,
                dynamic_ncols=True,
            ),
            **kwargs,
        )

        ds = open_zarr(output_subdir, consolidated=False).expand_dims(
            dim="chain",
            axis=0,
        )
        ds = ds.assign_coords(
            {k: np.arange(ds.sizes[k]) for k in ds.sizes},
        ).assign_attrs(_make_attrs())
        out = xr.DataTree(name="root")
        out["prior_predictive"] = xr.DataTree(ds)
        out["prior_predictive"].attrs["output_dir"] = str(output_subdir)
        if self.posterior:
            out["posterior"] = _dict_to_datatree(self.posterior)
        out.attrs["output_dir"] = str(output_dir)

        return out

    def sample(
        self,
        *,
        num_samples: int = 1000,
        rng_key: ArrayLike | None = None,
        return_sites: str | Iterable[str] | None = None,
        return_datatree: bool = True,
        **kwargs: object,
    ) -> xr.DataTree | dict[str, npt.NDArray]:
        """Draw posterior samples from a fitted model.

        Args:
            num_samples: The number of posterior samples to draw.
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed. Ignored if the
                inference method is MCMC, where the ``post_warmup_state`` property will
                be used to continue sampling.
            return_sites: Names of variables (sites) to return. If ``None``, samples
                all latent sites. Ignored if the inference method is MCMC.
            return_datatree: If ``True``, return a :external:class:`~xarray.DataTree`;
                otherwise return a :class:`dict`.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays. Only relevant when the inference method
                is MCMC.

        Returns:
            Posterior samples.

        Raises:
            TypeError: If :attr:`~aimz.ImpactModel.param_output` is not passed as an
                argument when the inference method is MCMC.
        """
        _check_is_fitted(self)

        if rng_key is None:
            self._rng_key, rng_key = random.split(self._rng_key)

        if isinstance(self.inference, MCMC):
            # Validate the provided parameters against the kernel's signature
            args_bound = signature(self.kernel).bind(**kwargs).arguments
            if self.param_output not in args_bound:
                msg = f"{self.param_output!r} must be provided in `.sample()`."
                raise TypeError(msg)
            self.inference.post_warmup_state = self.inference.last_state
            self.inference.num_samples = num_samples
            self.inference.run(self.inference.post_warmup_state.rng_key, **args_bound)
            posterior_samples = device_get(self.inference.get_samples())
        else:
            posterior_samples = device_get(
                _sample_forward(
                    substitute(
                        self.inference.guide,
                        data=cast("SVIRunResult", self.vi_result).params,
                    ),
                    rng_key=rng_key,
                    num_samples=num_samples,
                    return_sites=self._coerce_return_sites(return_sites)
                    if return_sites
                    else None,
                    samples=None,
                    model_kwargs=None,
                ),
            )

        if not return_datatree:
            return posterior_samples

        out = xr.DataTree(name="root")
        out["posterior"] = _dict_to_datatree(posterior_samples)

        return out

    def sample_posterior_predictive_on_batch(
        self,
        X: ArrayLike,
        *,
        intervention: dict | None = None,
        rng_key: ArrayLike | None = None,
        return_sites: str | Iterable[str] | None = None,
        return_datatree: bool = True,
        **kwargs: object,
    ) -> xr.DataTree | dict[str, npt.NDArray]:
        """Draw samples from the posterior predictive distribution.

        This method is a convenience alias for
        :meth:`~aimz.ImpactModel.predict_on_batch`, with ``in_sample`` automatically
        set to ``True``.

        Args:
            X (ArrayLike): Input data with shape ``(n_samples_X, n_features)``.
            intervention: A dictionary mapping sample sites to their corresponding
                intervention values. Interventions enable counterfactual analysis by
                modifying the specified sample sites during prediction (posterior
                predictive sampling).
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed.
            return_sites: Names of variables (sites) to return. If ``None``, samples
                :attr:`~aimz.ImpactModel.param_output` and deterministic sites.
            return_datatree: If ``True``, return a :external:class:`~xarray.DataTree`;
                otherwise return a :class:`dict`.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            Posterior predictive samples. Posterior samples are included if available.

        Raises:
            TypeError: If :attr:`~aimz.ImpactModel.param_output` is passed as an
                argument.

        See Also:
            :meth:`~aimz.ImpactModel.predict_on_batch`.
        """
        return self.predict_on_batch(
            X,
            intervention=intervention,
            rng_key=rng_key,
            in_sample=True,
            return_sites=return_sites,
            return_datatree=return_datatree,
            **kwargs,
        )

    def sample_posterior_predictive(
        self,
        X: ArrayLike | ArrayLoader,
        *,
        intervention: dict | None = None,
        rng_key: ArrayLike | None = None,
        return_sites: str | Iterable[str] | None = None,
        batch_size: int | None = None,
        output_dir: str | Path | None = None,
        progress: bool = True,
        **kwargs: object,
    ) -> xr.DataTree:
        """Draw samples from the posterior predictive distribution.

        This method is a convenience alias for :meth:`~aimz.ImpactModel.predict`, with
        ``in_sample`` automatically set to ``True``.

        Args:
            X (ArrayLike | ArrayLoader): Input data, either an array-like of shape
                ``(n_samples, n_features)`` or a data loader that holds all array-like
                objects and handles batching internally.
            intervention: A dictionary mapping sample sites to their corresponding
                intervention values. Interventions enable counterfactual analysis by
                modifying the specified sample sites during prediction (posterior
                predictive sampling).
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed.
            return_sites: Names of variables (sites) to return. If ``None``, samples
                :attr:`~aimz.ImpactModel.param_output` and deterministic sites.
            batch_size: The batch size for data loading during posterior predictive
                sampling. It also determines the chunk size used to store the samples.
                If ``None``, it is determined automatically based on the input data and
                number of samples. Ignored if ``X`` is a data loader, in which case the
                data loader is expected to handle batching internally.
            output_dir: The directory where the outputs will be saved. If the specified
                directory does not exist, it will be created automatically. If ``None``,
                a default temporary directory will be created. A timestamped
                subdirectory will be generated within this directory to store the
                outputs. Outputs are saved in the Zarr format.
            progress: Whether to display a progress bar.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            Posterior predictive samples. Posterior samples are included if available.

        Raises:
            TypeError: If :attr:`~aimz.ImpactModel.param_output` is passed as an
                argument.

        See Also:
            :meth:`~aimz.ImpactModel.predict()`.
        """
        return self.predict(
            X,
            intervention=intervention,
            rng_key=rng_key,
            in_sample=True,
            return_sites=return_sites,
            batch_size=batch_size,
            output_dir=output_dir,
            progress=progress,
            **kwargs,
        )

    def train_on_batch(
        self,
        X: ArrayLike,
        y: ArrayLike,
        rng_key: ArrayLike | None = None,
        **kwargs: object,
    ) -> tuple[SVIState, Array]:
        """Run a single VI step on the given batch of data.

        Args:
            X (ArrayLike): Input data with shape ``(n_samples_X, n_features)``.
            y (ArrayLike): Output data with shape ``(n_samples_Y,)``.
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed. The key is only
                used for initialization if the internal SVI state is not yet set.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            - Updated SVI state after the training step.

            - Loss value as a scalar array.

        Note:
            This method updates the internal SVI state on every call, so it is not
            necessary to capture the returned state externally unless explicitly needed.
            However, the returned loss value can be used for monitoring or logging.
        """
        batch = {self.param_input: X, self.param_output: y, **kwargs}

        if self._vi_state is None:
            self._build_kernel_spec(
                signature(self.kernel).bind(**batch).arguments,
                with_output=True,
            )
            if rng_key is None:
                self._rng_key, rng_key = random.split(self._rng_key)
            self._vi_state = self.inference.init(rng_key, **batch)
        if self._fn_vi_update is None:
            _, kwargs_extra = _group_kwargs(kwargs)
            self._fn_vi_update = jit(
                self.inference.update,
                static_argnames=tuple(kwargs_extra._fields),
            )

        self._vi_state, loss = self._fn_vi_update(self._vi_state, **batch)

        return self._vi_state, loss

    def fit_on_batch(
        self,
        X: ArrayLike,
        y: ArrayLike,
        *,
        num_steps: int = 10000,
        num_samples: int = 1000,
        rng_key: ArrayLike | None = None,
        progress: bool = True,
        **kwargs: object,
    ) -> Self:
        """Fit the impact model to the provided batch of data.

        This method behaves differently depending on the inference method specified at
        initialization of the ImpactModel instance:

        - SVI
            Runs variational inference on the provided batch by invoking the
            :external:meth:`~numpyro.infer.svi.SVI.run` method of the
            :external:class:`~numpyro.infer.svi.SVI` instance from NumPyro to
            estimate the posterior distribution, then draws samples from it.

        - MCMC
            Runs posterior sampling by invoking the
            :external:meth:`~numpyro.infer.mcmc.MCMC.run` method of the
            :external:class:`~numpyro.infer.mcmc.MCMC` instance from NumPyro.

        Args:
            X (ArrayLike): Input data with shape ``(n_samples_X, n_features)``.
            y (ArrayLike): Output data with shape ``(n_samples_Y,)``.
            num_steps: Number of steps for variational inference optimization. Ignored
                if the inference method is MCMC.
            num_samples: The number of posterior samples to draw. Ignored if the
                inference method is MCMC.
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed.
            progress: Whether to display a progress bar. Ignored if the inference method
                is MCMC.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            The fitted model instance, enabling method chaining.

        Note:
            This method continues training from the existing SVI state if available. To
            start training from scratch, create a new model instance.
        """
        X, y = _validate_X_y_to_jax(X, y)

        # Validate the provided parameters against the kernel's signature
        args_bound = (
            signature(self.kernel)
            .bind(**{self.param_input: X, self.param_output: y, **kwargs})
            .arguments
        )
        self._build_kernel_spec(args_bound, with_output=True)
        if rng_key is None:
            self._rng_key, rng_key = random.split(self._rng_key)
        rng_key, rng_subkey = random.split(rng_key)
        if isinstance(self.inference, SVI):
            self._num_samples = num_samples
            logger.info("Performing variational inference optimization...")
            self.vi_result = self.inference.run(
                rng_subkey,
                num_steps=num_steps,
                progress_bar=progress,
                init_state=self._vi_state,
                **args_bound,
            )
            self._vi_state = self.vi_result.state
            if np.any(np.isnan(self.vi_result.losses)):
                warn(
                    "Loss contains NaN or Inf, indicating numerical instability.",
                    category=RuntimeWarning,
                    stacklevel=2,
                )
            logger.info("Posterior sampling...")
            rng_key, rng_subkey = random.split(rng_key)
            self._posterior = _sample_forward(
                substitute(self.inference.guide, data=self.vi_result.params),
                rng_key=rng_subkey,
                num_samples=self._num_samples,
                return_sites=None,
                samples=None,
                model_kwargs=None,
            )
        elif isinstance(self.inference, MCMC):
            logger.info("Posterior sampling...")
            self.inference.run(rng_subkey, **args_bound)
            self._posterior = device_get(self.inference.get_samples())
            self._num_samples = (
                next(iter(self.posterior.values())).shape[0] if self.posterior else 0
            )

        self._is_fitted = True

        return self

    def fit(
        self,
        X: ArrayLike | ArrayLoader,
        y: ArrayLike | None = None,
        *,
        num_samples: int = 1000,
        rng_key: ArrayLike | None = None,
        progress: bool = True,
        batch_size: int | None = None,
        epochs: int = 1,
        shuffle: bool = True,
        **kwargs: object,
    ) -> Self:
        """Fit the impact model to the provided data using epoch-based training.

        This method implements an epoch-based training loop, where the data is iterated
        over in minibatches for a specified number of epochs. Variational inference is
        performed by repeatedly updating the model parameters on each minibatch, and
        then posterior samples are drawn from the fitted model.

        Args:
            X (ArrayLike | ArrayLoader): Input data, either an array-like of shape
                ``(n_samples, n_features)`` or a data loader that holds all array-like
                objects and handles batching internally.
            y (ArrayLike | None): Output data with shape ``(n_samples_Y,)``. Must be
                ``None`` if ``X`` is a data loader.
            num_samples: The number of posterior samples to draw.
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed.
            progress: Whether to display a progress bar.
            batch_size: The number of data points processed at each step of variational
                inference. If ``None``, the entire dataset is used as a single batch in
                each epoch. Ignored if ``X`` is a data loader, in which case the data
                loader is expected to handle batching internally.
            epochs: The number of epochs for variational inference optimization.
            shuffle: Whether to shuffle the data at each epoch. Ignored if ``X`` is a
                data loader.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            The fitted model instance, enabling method chaining.

        Raises:
            TypeError: If the inference method is MCMC.

        Note:
            This method continues training from the existing SVI state if available.
            To start training from scratch, create a new model instance. It does not
            check whether the model or guide is written to support subsampling semantics
            (e.g., using NumPyro's :external:func:`~numpyro.primitives.subsample` or
            similar constructs).
        """
        if isinstance(self.inference, MCMC):
            msg = (
                "`.fit()` is not supported for MCMC inference. Use `.fit_on_batch()` "
                "instead."
            )
            raise TypeError(msg)

        self._num_samples = num_samples

        if rng_key is None:
            self._rng_key, rng_key = random.split(self._rng_key)

        rng_key, rng_subkey = random.split(rng_key)
        dataloader, kwargs_extra = _setup_inputs(
            X=X,
            y=y,
            rng_key=rng_subkey,
            batch_size=batch_size if batch_size is not None else len(cast("Sized", X)),
            num_samples=num_samples,
            shuffle=shuffle,
            device=None,
            **kwargs,
        )

        logger.info("Performing variational inference optimization...")
        losses: list[float] = []
        rng_key, rng_subkey = random.split(rng_key)
        for epoch in range(epochs):
            losses_epoch: list[float] = []
            pbar = tqdm(
                dataloader,
                desc=f"Epoch {epoch + 1}/{epochs}",
                total=len(dataloader),
                disable=not progress,
                dynamic_ncols=True,
            )
            for batch, _ in pbar:
                self._vi_state, loss = self.train_on_batch(
                    **batch,
                    **kwargs_extra._asdict(),
                    rng_key=rng_subkey,
                )
                loss_batch = device_get(loss)
                losses_epoch.append(loss_batch)
                pbar.set_postfix({"loss": f"{float(loss_batch):.4f}"})
            losses_epoch_arr = jnp.stack(losses_epoch)
            losses.extend(losses_epoch_arr)
            tqdm.write(
                f"Epoch {epoch + 1}/{epochs} - "
                f"Average loss: {float(jnp.mean(losses_epoch_arr)):.4f}",
            )
        self.vi_result = SVIRunResult(
            params=self.inference.get_params(self._vi_state),
            state=self._vi_state,
            losses=jnp.asarray(losses),
        )
        if np.any(np.isnan(self.vi_result.losses)):
            msg = "Loss contains NaN or Inf, indicating numerical instability."
            warn(msg, category=RuntimeWarning, stacklevel=2)

        self._is_fitted = True

        logger.info("Posterior sampling...")
        rng_key, rng_subkey = random.split(rng_key)
        self._posterior = _sample_forward(
            substitute(self.inference.guide, data=self.vi_result.params),
            rng_key=rng_subkey,
            num_samples=self._num_samples,
            return_sites=None,
            samples=None,
            model_kwargs=None,
        )

        return self

    def is_fitted(self) -> bool:
        """Check fitted status.

        Returns:
            `True` if the model is fitted, `False` otherwise.

        """
        return hasattr(self, "_is_fitted") and self._is_fitted

    def set_posterior_sample(self, posterior_sample: dict[str, Array]) -> Self:
        """Set posterior samples for the model.

        This method sets externally obtained posterior samples on the model instance,
        enabling downstream analysis without requiring a call to
        :meth:`~aimz.ImpactModel.fit` or :meth:`~aimz.ImpactModel.fit_on_batch`.

        It is primarily intended for workflows where posterior sampling is performed
        manually—for example, using NumPyro's :external:class:`~numpyro.infer.svi.SVI`
        (or :external:class:`~numpyro.infer.mcmc.MCMC`) with the
        :external:class:`~numpyro.infer.util.Predictive` API—and the resulting
        posterior samples are injected into the model for further use.

        Internally, ``batch_ndims`` is set to ``1`` by default to correctly handle the
        batch dimensions of the posterior samples. For more information, refer to the
        `NumPyro documentation <https://num.pyro.ai/en/stable/utilities.html#predictive>`__.

        Args:
            posterior_sample: Posterior samples to set for the model.

        Returns:
            The model instance, treated as fitted with posterior samples set, enabling
            method chaining.

        Raises:
            ValueError: If the batch shapes in ``posterior_sample`` are inconsistent.
        """
        batch_ndims = 1
        batch_shapes = {
            sample.shape[:batch_ndims] for sample in posterior_sample.values()
        }
        if not batch_shapes:
            msg = "`posterior_sample` cannot be empty."
            raise ValueError(msg)
        if len(batch_shapes) > 1:
            msg = (
                f"Inconsistent batch shapes found in `posterior_sample`: {batch_shapes}"
            )
            raise ValueError(msg)
        (self._num_samples,) = batch_shapes.pop()
        self._posterior = posterior_sample
        if self._kernel_spec is None:
            self._kernel_spec = KernelSpec(
                traced=False,
                sample_sites=tuple(cast("dict[str, Array]", self._posterior).keys()),
                return_sites=(self.param_output,),
                output_observed=False,
            )
        self._is_fitted = True

        return self

    def predict_on_batch(
        self,
        X: ArrayLike,
        *,
        intervention: dict | None = None,
        rng_key: ArrayLike | None = None,
        in_sample: bool = True,
        return_sites: str | Iterable[str] | None = None,
        return_datatree: bool = True,
        **kwargs: object,
    ) -> xr.DataTree | dict[str, npt.NDArray]:
        """Predict the output based on the fitted model.

        This method returns predictions for a single batch of input data and is better
        suited for:

            1) Models incompatible with :meth:`~aimz.ImpactModel.predict` due to their
            posterior sample shapes.

            2) Scenarios where writing results to to files (e.g., disk, cloud storage)
            is not desired.

            3) Smaller datasets, as this method may be slower due to limited
            parallelism.

        Args:
            X (ArrayLike): Input data with shape ``(n_samples_X, n_features)``.
            intervention: A dictionary mapping sample sites to their corresponding
                intervention values. Interventions enable counterfactual analysis by
                modifying the specified sample sites during prediction (posterior
                predictive sampling).
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed.
            in_sample: Specifies the group where posterior predictive samples are stored
                in the returned output. If ``True``, samples are stored in the
                ``posterior_predictive`` group, indicating they were generated based on
                data used during model fitting. If ``False``, samples are stored in the
                ``predictions`` group, indicating they were generated based on
                out-of-sample data.
            return_sites: Names of variables (sites) to return. If ``None``, samples
                :attr:`~aimz.ImpactModel.param_output` and deterministic sites.
            return_datatree: If ``True``, return a :external:class:`~xarray.DataTree`;
                otherwise return a :class:`dict`.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            Posterior predictive samples. Posterior samples are included if available.

        Raises:
            TypeError: If :attr:`~aimz.ImpactModel.param_output` is passed as an
                argument.
        """
        _check_is_fitted(self)

        X = cast("Array", _validate_X_y_to_jax(X))

        # Validate the provided parameters against the kernel's signature
        args_bound = (
            signature(self.kernel).bind(**{self.param_input: X, **kwargs}).arguments
        )
        if self.param_output in args_bound:
            msg = f"Specifying {self.param_output!r} is not allowed."
            raise TypeError(msg)

        if rng_key is None:
            self._rng_key, rng_key = random.split(self._rng_key)

        if intervention is None:
            kernel = self.kernel
        else:
            rng_key, rng_subkey = random.split(rng_key)
            kernel = seed(do(self.kernel, data=intervention), rng_seed=rng_subkey)

        samples = device_get(
            _sample_forward(
                kernel,
                rng_key=rng_key,
                num_samples=self._num_samples,
                return_sites=self._coerce_return_sites(return_sites),
                samples=self.posterior,
                model_kwargs=args_bound,
            ),
        )

        if not return_datatree:
            return samples

        out = xr.DataTree(name="root")
        if self.posterior:
            out["posterior"] = _dict_to_datatree(self.posterior)
        out["posterior_predictive" if in_sample else "predictions"] = _dict_to_datatree(
            samples,
        )

        return out

    def predict(
        self,
        X: ArrayLike | ArrayLoader,
        *,
        intervention: dict | None = None,
        rng_key: ArrayLike | None = None,
        in_sample: bool = True,
        return_sites: str | Iterable[str] | None = None,
        batch_size: int | None = None,
        output_dir: str | Path | None = None,
        progress: bool = True,
        **kwargs: object,
    ) -> xr.DataTree:
        """Predict the output based on the fitted model.

        This method performs posterior predictive sampling to generate model-based
        predictions. It is optimized for batch processing of large input data and is not
        recommended for use in loops that process only a few inputs at a time. Results
        are written to disk in the Zarr format, with sampling and file writing decoupled
        and executed concurrently.

        Args:
            X (ArrayLike | ArrayLoader): Input data, either an array-like of shape
                ``(n_samples, n_features)`` or a data loader that holds all array-like
                objects and handles batching internally.
            intervention: A dictionary mapping sample sites to their corresponding
                intervention values. Interventions enable counterfactual analysis by
                modifying the specified sample sites during prediction (posterior
                predictive sampling).
            rng_key (ArrayLike | None): A pseudo-random number generator key. By
                default, an internal key is used and split as needed.
            in_sample: Specifies the group where posterior predictive samples are stored
                in the returned output. If ``True``, samples are stored in the
                ``posterior_predictive`` group, indicating they were generated based on
                data used during model fitting. If ``False``, samples are stored in the
                ``predictions`` group, indicating they were generated based on
                out-of-sample data.
            return_sites: Names of variables (sites) to return. If ``None``, samples
                :attr:`~aimz.ImpactModel.param_output` and deterministic sites.
            batch_size: The batch size for data loading during posterior predictive
                sampling. It also determines the chunk size used to store the samples.
                If ``None``, it is determined automatically based on the input data
                and number of samples. Ignored if ``X`` is a data loader, in which case
                the data loader is expected to handle batching internally.
            output_dir: The directory where the outputs will be saved. If the specified
                directory does not exist, it will be created automatically. If ``None``,
                a default temporary directory will be created. A timestamped
                subdirectory will be generated within this directory to store the
                outputs. Outputs are saved in the Zarr format.
            progress: Whether to display a progress bar.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            Posterior predictive samples. Posterior samples are included if available.

        Raises:
            TypeError: If :attr:`~aimz.ImpactModel.param_output` is passed as an
                argument.

        See Also:
            :meth:`~aimz.ImpactModel.cleanup` to remove the temporary directory if
            created.
        """
        _check_is_fitted(self)

        if isinstance(X, ArrayLike):
            ndim_posterior_sample = 2
            if self.posterior and any(
                v.ndim == ndim_posterior_sample and v.shape[1] == len(cast("Sized", X))
                for v in self.posterior.values()
            ):
                msg = (
                    "One or more posterior sample shapes are not compatible with "
                    "`.predict()` under sharded parallelism; falling back to "
                    "`.predict_on_batch()`."
                )
                warn(msg, category=UserWarning, stacklevel=2)

                return cast(
                    "xr.DataTree",
                    self.predict_on_batch(
                        cast("ArrayLike", X),
                        intervention=intervention,
                        rng_key=rng_key,
                        in_sample=in_sample,
                        return_sites=return_sites
                        or cast("KernelSpec", self._kernel_spec).return_sites,
                        return_datatree=True,
                        **kwargs,
                    ),
                )
            # Validate the provided parameters against the kernel's signature
            args_bound = (
                signature(self.kernel).bind(**{self.param_input: X, **kwargs}).arguments
            )
            if self.param_output in args_bound:
                msg = f"Specifying {self.param_output!r} is not allowed."
                raise TypeError(msg)

        if rng_key is None:
            self._rng_key, rng_key = random.split(self._rng_key)

        if intervention is None:
            kernel = self.kernel
        else:
            rng_key, rng_subkey = random.split(rng_key)
            kernel = seed(do(self.kernel, data=intervention), rng_seed=rng_subkey)

        return_sites = self._coerce_return_sites(return_sites)

        kwargs_array, kwargs_extra = _group_kwargs(kwargs)
        if self._fn_sample_posterior_predictive is None:
            self._fn_sample_posterior_predictive = _create_sharded_sampler(
                self._mesh,
                len(kwargs_array),
                len(kwargs_extra),
            )

        output_dir, output_subdir = self._create_output_subdir(output_dir)

        dataloader, _ = _setup_inputs(
            X=X,
            y=None,
            rng_key=self.rng_key,
            batch_size=batch_size,
            num_samples=self._num_samples,
            shuffle=False,
            device=self._device,
            **kwargs,
        )

        self._sample_and_write(
            num_samples=self._num_samples,
            rng_key=rng_key,
            return_sites=return_sites,
            output_dir=output_subdir,
            kernel=kernel,
            sampler=self._fn_sample_posterior_predictive,
            samples=cast("dict[str, Array]", self.posterior),
            dataloader=dataloader,
            pbar=tqdm(
                desc=(f"Posterior predictive sampling [{', '.join(return_sites)}]"),
                total=len(dataloader),
                disable=not progress,
                dynamic_ncols=True,
            ),
            **kwargs,
        )

        ds = open_zarr(output_subdir, consolidated=False).expand_dims(
            dim="chain",
            axis=0,
        )
        ds = ds.assign_coords(
            {k: np.arange(ds.sizes[k]) for k in ds.sizes},
        ).assign_attrs(_make_attrs())
        out = xr.DataTree(name="root")
        if self.posterior:
            out["posterior"] = _dict_to_datatree(self.posterior)
        group = "posterior_predictive" if in_sample else "predictions"
        out[group] = xr.DataTree(ds)
        out[group].attrs["output_dir"] = str(output_subdir)
        out.attrs["output_dir"] = str(output_dir)

        return out

    def estimate_effect(
        self,
        output_baseline: xr.DataTree | None = None,
        output_intervention: xr.DataTree | None = None,
        args_baseline: dict | None = None,
        args_intervention: dict | None = None,
    ) -> xr.DataTree:
        """Estimate the effect of an intervention.

        .. _NumPyro: https://num.pyro.ai/

        This computes (intervention - baseline) for every variable in the shared
        predictive group, preserving sampling (chain/draw) dimensions. When
        interventions are used in prediction they are applied internally through
        NumPyro_'s :external:class:`~numpyro.handlers.do` effect handler (graph surgery)
        without requiring model rewrites.

        Args:
            output_baseline: Precomputed output for the baseline scenario.
            output_intervention: Precomputed output for the intervention scenario.
            args_baseline: Input arguments for the baseline scenario. Passed to the
                :meth:`~aimz.ImpactModel.predict` to compute predictions if
                ``output_baseline`` is not provided. Ignored if ``output_baseline`` is
                already given.
            args_intervention: Input arguments for the intervention scenario. Passed to
                the :meth:`~aimz.ImpactModel.predict` to compute predictions if
                ``output_intervention`` is not provided. Ignored if
                ``output_intervention`` is already given.

        Returns:
            The estimated impact of an intervention. Posterior samples are included if
            available.

        Raises:
            ValueError: If neither ``output_baseline`` nor ``args_baseline`` is
                provided, or if neither ``output_intervention`` nor
                ``args_intervention`` is provided.

        See Also:
            :meth:`~aimz.ImpactModel.cleanup` to remove the temporary directory if
            created.
        """
        _check_is_fitted(self)

        if output_baseline:
            dt_baseline = output_baseline
        elif args_baseline:
            dt_baseline = self.predict(**args_baseline)
        else:
            msg = "Either `output_baseline` or `args_baseline` must be provided."
            raise ValueError(msg)

        if output_intervention:
            dt_intervention = output_intervention
        elif args_intervention:
            dt_intervention = self.predict(**args_intervention)
        else:
            msg = (
                "Either `output_intervention` or `args_intervention` must be provided."
            )
            raise ValueError(msg)

        group = _validate_group(dt_baseline, dt_intervention)

        out = xr.DataTree(name="root")
        out[group] = dt_intervention[group] - dt_baseline[group]
        if self.posterior:
            out["posterior"] = _dict_to_datatree(self.posterior)
        # Propagate an output directory attribute to the effect result.
        # Precedence:
        #   1) intervention output_dir (if present)
        #   2) baseline output_dir (if present)
        #   3) model temporary directory (if it exists)
        output_dir_attr = (
            dt_intervention.attrs.get("output_dir")
            or dt_baseline.attrs.get("output_dir")
            or (self.temp_dir if self.temp_dir is not None else None)
        )
        if output_dir_attr is not None:
            out.attrs["output_dir"] = output_dir_attr

        return out

    def log_likelihood(
        self,
        X: ArrayLike | ArrayLoader,
        y: ArrayLike | None = None,
        *,
        batch_size: int | None = None,
        output_dir: str | Path | None = None,
        progress: bool = True,
        **kwargs: object,
    ) -> xr.DataTree:
        """Compute the log-likelihood of the data under the given model.

        Results are written to disk in the Zarr format, with computing and file writing
        decoupled and executed concurrently.

        Args:
            X (ArrayLike | ArrayLoader): Input data, either an array-like of shape
                ``(n_samples, n_features)`` or a data loader that holds all array-like
                objects and handles batching internally.
            y (ArrayLike | None): Output data with shape ``(n_samples_Y,)``. Must be
                ``None`` if ``X`` is a data loader.
            batch_size: The batch size for data loading during log-likelihood
                computation. It also determines the chunk size used to store the
                samples. If ``None``, it is determined automatically based on the input
                data and number of samples. Ignored if ``X`` is a data loader, in which
                case the data loader is expected to handle batching internally.
            output_dir: The directory where the outputs will be saved. If the specified
                directory does not exist, it will be created automatically. If ``None``,
                a default temporary directory will be created. A timestamped
                subdirectory will be generated within this directory to store the
                outputs. Outputs are saved in the Zarr format.
            progress: Whether to display a progress bar.
            **kwargs: Additional arguments passed to the model. All array-like values
                are expected to be JAX arrays.

        Returns:
            Log-likelihood values. Posterior samples are included if available.

        See Also:
            :meth:`~aimz.ImpactModel.cleanup` to remove the temporary directory if
            created.
        """
        _check_is_fitted(self)

        kwargs_array, kwargs_extra = _group_kwargs(kwargs)
        if self._fn_log_likelihood is None:
            self._fn_log_likelihood = _create_sharded_log_likelihood(
                self._mesh,
                len(kwargs_array),
                len(kwargs_extra),
            )

        output_dir, output_subdir = self._create_output_subdir(output_dir)

        dataloader, _ = _setup_inputs(
            X=X,
            y=y,
            rng_key=self.rng_key,
            batch_size=batch_size,
            num_samples=self._num_samples,
            shuffle=False,
            device=self._device,
            **kwargs,
        )

        site = self.param_output
        pbar = tqdm(
            desc=(f"Computing log-likelihood of {site}..."),
            total=len(dataloader),
            disable=not progress,
            dynamic_ncols=True,
        )

        zarr_group = open_group(output_subdir, mode="w")
        zarr_arr = {}
        threads, queues, error_queue = _start_writer_threads(
            (site,),
            group_path=output_subdir,
            writer=_writer,
            queue_size=min(cpu_count() or 1, 4),
        )
        try:
            for batch, n_pad in dataloader:
                kwargs_batch = [
                    v
                    for k, v in batch.items()
                    if k not in (self.param_input, self.param_output)
                ]
                arr = device_get(
                    self._fn_log_likelihood(
                        # Although computing the log-likelihood is deterministic, the
                        # model still needs to be seeded in order to trace its graph.
                        seed(self.kernel, rng_seed=self.rng_key),
                        self.posterior,
                        self.param_input,
                        site,
                        kwargs_array._fields + kwargs_extra._fields,
                        batch[self.param_input],
                        batch[self.param_output],
                        *(*kwargs_batch, *kwargs_extra),
                    ),
                )
                if site not in zarr_arr:
                    draws = self._num_samples if self.posterior else 1
                    zarr_arr[site] = zarr_group.create_array(
                        name=site,
                        shape=(draws, 0, *arr.shape[2:]),
                        dtype="float32" if arr.dtype == "bfloat16" else arr.dtype,
                        chunks=(draws, dataloader.batch_size, *arr.shape[2:]),
                        dimension_names=(
                            "draw",
                            *tuple(f"{site}_dim_{i}" for i in range(arr.ndim - 1)),
                        ),
                    )
                queues[site].put(arr[:, : -n_pad or None])
                if not error_queue.empty():
                    _, exc, tb = error_queue.get()
                    raise exc.with_traceback(tb)
                pbar.update()
            pbar.set_description("Computation complete, writing in progress...")
            _shutdown_writer_threads(threads, queues=queues)
        except:
            _shutdown_writer_threads(threads, queues=queues)
            logger.exception(
                "Exception encountered. Cleaning up output directory: %s",
                output_subdir,
            )
            rmtree(output_subdir, ignore_errors=True)
            raise
        finally:
            pbar.close()

        ds = open_zarr(output_subdir, consolidated=False).expand_dims(
            dim="chain",
            axis=0,
        )
        ds = ds.assign_coords(
            {k: np.arange(ds.sizes[k]) for k in ds.sizes},
        ).assign_attrs(_make_attrs())
        out = xr.DataTree(name="root")
        if self.posterior:
            out["posterior"] = _dict_to_datatree(self.posterior)
        out["log_likelihood"] = xr.DataTree(ds)
        out["log_likelihood"].attrs["output_dir"] = str(output_subdir)
        out.attrs["output_dir"] = str(output_dir)

        return out

    def cleanup(self) -> None:
        """Clean up the temporary directory created for storing outputs.

        If the temporary directory was never created or has already been cleaned up,
        this method does nothing. It does not delete any explicitly specified output
        directory. While the temporary directory is typically removed automatically
        during garbage collection, this behavior is not guaranteed. Therefore, calling
        this method explicitly is recommended to ensure timely resource release.
        """
        if hasattr(self, "_temp_dir") and self._temp_dir is not None:
            logger.info("Temporary directory cleaned up at: %s", self._temp_dir.name)
            self._temp_dir.cleanup()
            self._temp_dir = None



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MLflow model saving for aimz."""

from __future__ import annotations

import tempfile
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING

import yaml
from mlflow import pyfunc
from mlflow.models import (
    Model,
    ModelInputExample,
    ModelSignature,
    infer_pip_requirements,
    infer_signature,
)
from mlflow.models.model import MLMODEL_FILE_NAME
from mlflow.models.utils import _save_example
from mlflow.utils import _get_fully_qualified_class_name
from mlflow.utils.autologging_utils import (
    INPUT_EXAMPLE_SAMPLE_ROWS,
)
from mlflow.utils.docstring_utils import LOG_MODEL_PARAM_DOCS, format_docstring
from mlflow.utils.environment import (
    _CONDA_ENV_FILE_NAME,
    _CONSTRAINTS_FILE_NAME,
    _PYTHON_ENV_FILE_NAME,
    _REQUIREMENTS_FILE_NAME,
    _mlflow_conda_env,
    _process_conda_env,
    _process_pip_requirements,
    _PythonEnv,
    _validate_env_arguments,
)
from mlflow.utils.file_utils import get_total_file_size, write_to
from mlflow.utils.model_utils import (
    _validate_and_copy_code_paths,
    _validate_and_prepare_target_save_path,
)
from mlflow.utils.requirements_utils import _get_pinned_requirement

if TYPE_CHECKING:
    from collections.abc import Iterable

    from aimz.model.impact_model import ImpactModel
FLAVOR_NAME = "aimz"


def get_default_conda_env(*, include_cloudpickle: bool = False) -> dict[str, object]:
    """Return the default Conda environment for saving an aimz model.

    Args:
        include_cloudpickle: If ``True``, include ``cloudpickle`` in the environment.

    Returns:
        The default Conda environment for MLflow Models produced by calls to
        :func:`~aimz.mlflow.save_model` and :func:`~aimz.mlflow.log_model`.
    """
    return _mlflow_conda_env(
        additional_pip_deps=get_default_pip_requirements(
            include_cloudpickle=include_cloudpickle,
        ),
    )


def get_default_pip_requirements(*, include_cloudpickle: bool = False) -> list[str]:
    """Return the default pip requirements for saving an aimz model.

    Args:
        include_cloudpickle: If ``True``, include ``cloudpickle`` in the requirements
            list.

    Returns:
        The default pip requirements for MLflow Models produced by calls to
        :func:`~aimz.mlflow.save_model` and :func:`~aimz.mlflow.log_model`.
    """
    out = [_get_pinned_requirement("aimz")]
    if include_cloudpickle:
        out.append(_get_pinned_requirement("cloudpickle"))

    return out


@format_docstring(LOG_MODEL_PARAM_DOCS.format(package_name=FLAVOR_NAME))
def save_model(
    model: ImpactModel,
    path: str | Path,
    conda_env: dict | None = None,
    code_paths: list | None = None,
    mlflow_model: Model | None = None,
    signature: ModelSignature | None = None,
    input_example: ModelInputExample = None,
    pip_requirements: Iterable[str] | str | None = None,
    extra_pip_requirements: Iterable[str] | str | None = None,
    metadata: dict | None = None,
) -> None:
    """Save an aimz model to a path on the local file system.

    Args:
        model: An aimz model (an instance of :class:`~aimz.ImpactModel`) to be saved.
        path (str | Path): Local path where the model is to be saved.
        conda_env: {{ conda_env }}
        code_paths: {{ code_paths }}
        mlflow_model: :class:`mlflow.models.Model` this flavor is being added to.
        signature: {{ signature }}
        input_example: {{ input_example }}
        pip_requirements: {{ pip_requirements }}
        extra_pip_requirements: {{ extra_pip_requirements }}
        metadata: {{ metadata }}
    """
    import cloudpickle

    _validate_env_arguments(conda_env, pip_requirements, extra_pip_requirements)

    path = Path(path).resolve()
    _validate_and_prepare_target_save_path(path)
    model_data_subpath = "model.pkl"
    model_data_path = path / model_data_subpath
    code_dir_subpath = _validate_and_copy_code_paths(code_paths, path)

    with model_data_path.open("wb") as out:
        cloudpickle.dump(model, out)

    if mlflow_model is None:
        mlflow_model = Model()
    saved_example = _save_example(
        mlflow_model,
        input_example=input_example,
        path=str(path),
    )
    if signature is None and saved_example is not None:
        with tempfile.TemporaryDirectory() as temp_dir:
            signature = infer_signature(
                input_example,
                params={
                    "batch_size": INPUT_EXAMPLE_SAMPLE_ROWS,
                    "output_dir": temp_dir,
                    "progress": False,
                },
            )
    if signature is not None:
        mlflow_model.signature = signature
    if metadata is not None:
        mlflow_model.metadata = metadata

    flavor_options = {
        "version": version("aimz"),
        "model_class": _get_fully_qualified_class_name(model),
        "serialization_format": "cloudpickle",
        "pickled_model": model_data_subpath,
    }
    mlflow_model.add_flavor(FLAVOR_NAME, code=code_dir_subpath, **flavor_options)
    pyfunc.add_to_model(
        mlflow_model,
        loader_module="aimz.mlflow.pyfunc",
        data=model_data_subpath,
        conda_env=_CONDA_ENV_FILE_NAME,
        python_env=_PYTHON_ENV_FILE_NAME,
        code=code_dir_subpath,
    )
    if size := get_total_file_size(path):
        mlflow_model.model_size_bytes = size
    mlflow_model.save(path / MLMODEL_FILE_NAME)

    if conda_env is None:
        if pip_requirements is None:
            default_reqs = get_default_pip_requirements(include_cloudpickle=True)
            # To ensure `_load_pyfunc` can successfully load the model during the
            # dependency inference, `mlflow_model.save` must be called beforehand to
            # save an MLmodel file.
            inferred_reqs = infer_pip_requirements(
                path,
                flavor=FLAVOR_NAME,
                fallback=default_reqs,
            )
            default_reqs = sorted(set(inferred_reqs).union(default_reqs))
        else:
            default_reqs = None
        conda_env, pip_requirements, pip_constraints = _process_pip_requirements(
            default_reqs,
            pip_requirements,
            extra_pip_requirements,
        )
    else:
        conda_env, pip_requirements, pip_constraints = _process_conda_env(conda_env)

    with Path.open(path / _CONDA_ENV_FILE_NAME, "w") as f:
        yaml.safe_dump(conda_env, stream=f, default_flow_style=False)

    # Save `constraints.txt` if necessary.
    if pip_constraints:
        write_to(path / _CONSTRAINTS_FILE_NAME, "\n".join(pip_constraints))

    # Save `requirements.txt`.
    write_to(path / _REQUIREMENTS_FILE_NAME, "\n".join(pip_requirements))

    _PythonEnv.current().to_yaml(path / _PYTHON_ENV_FILE_NAME)



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the `.sample_prior_predictive_on_batch()` method."""

import jax.numpy as jnp
import numpyro.distributions as dist
import pytest
import xarray as xr
from conftest import lm
from jax import random
from jax.typing import ArrayLike
from numpyro import sample
from numpyro.infer import SVI, Trace_ELBO
from numpyro.infer.autoguide import AutoNormal
from numpyro.optim import Adam

from aimz import ImpactModel
from aimz._exceptions import KernelValidationError


def test_kernel_without_output(synthetic_data: tuple[ArrayLike, ArrayLike]) -> None:
    """Kernel without output sample site raises an error."""

    def kernel(X: ArrayLike, y: ArrayLike | None = None) -> None:
        sample("z", dist.Delta(y if y is not None else jnp.zeros(len(X))), obs=y)

    X, _ = synthetic_data
    im = ImpactModel(
        kernel,
        rng_key=random.key(42),
        inference=SVI(
            kernel,
            guide=AutoNormal(kernel),
            optim=Adam(step_size=1e-3),
            loss=Trace_ELBO(),
        ),
    )

    with pytest.raises(KernelValidationError):
        im.sample_prior_predictive_on_batch(X)


@pytest.mark.parametrize("vi", [lm], indirect=True)
class TestKernelParameterValidation:
    """Test class for validating parameter compatibility with the kernel."""

    def test_invalid_parameter(
        self,
        synthetic_data: tuple[ArrayLike, ArrayLike],
        vi: "SVI",
    ) -> None:
        """An invalid parameter raise an error."""
        X, y = synthetic_data
        im = ImpactModel(lm, rng_key=random.key(42), inference=vi)
        with pytest.raises(TypeError):
            im.sample_prior_predictive_on_batch(X=X, y=y)


@pytest.mark.parametrize("vi", [lm], indirect=True)
def test_sample_prior_predictive_on_batch_lm(
    synthetic_data: tuple[ArrayLike, ArrayLike],
    vi: "SVI",
) -> None:
    """Test the `.sample_prior_predictive_on_batch()` method of ImpactModel."""
    X, y = synthetic_data
    im = ImpactModel(lm, rng_key=random.key(42), inference=vi)
    im.fit_on_batch(X, y)
    samples = im.sample_prior_predictive_on_batch(X=X, num_samples=99, return_sites="y")

    assert isinstance(samples, xr.DataTree)
    assert samples.prior_predictive["y"].values.shape == (1, 99, len(X))

    samples_dict = im.sample_prior_predictive_on_batch(
        X=X,
        num_samples=99,
        return_datatree=False,
        return_sites=["b", "y", "sigma"],
    )

    assert isinstance(samples_dict, dict)
    assert samples_dict["y"].shape == (99, len(X))
    assert im.kernel_spec.traced
    assert im.kernel_spec.output_observed



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the `.sample()` method."""

from typing import TYPE_CHECKING

import pytest
from conftest import lm
from jax import random
from jax.typing import ArrayLike
from numpyro.infer import MCMC

from aimz import ImpactModel

if TYPE_CHECKING:
    from numpyro.infer import SVI


@pytest.mark.parametrize("mcmc", [lm], indirect=True)
def test_missing_param_output(
    synthetic_data: tuple[ArrayLike, ArrayLike],
    mcmc: MCMC,
) -> None:
    """Missing `param_output` argument raises TypeError."""
    X, y = synthetic_data
    im = ImpactModel(lm, rng_key=random.key(42), inference=mcmc)
    im.fit_on_batch(X=X, y=y)
    with pytest.raises(TypeError):
        im.sample(rng_key=random.key(42), X=X)


@pytest.mark.parametrize("vi", [lm], indirect=True)
def test_sample_with_vi(
    synthetic_data: tuple[ArrayLike, ArrayLike],
    vi: "SVI",
) -> None:
    """Test the `.sample()` method of ImpactModel with SVI."""
    X, y = synthetic_data
    im = ImpactModel(lm, rng_key=random.key(42), inference=vi)
    im.fit_on_batch(X=X, y=y)

    num_samples = 7
    samples = im.sample(
        num_samples=num_samples,
        rng_key=random.key(42),
        return_sites="b",
        X=X,
        y=y,
    ).posterior

    # Check shapes for all sampled sites
    for var in samples.data_vars:
        assert samples[var].values.shape[1] == num_samples, (
            f"Incorrect number of samples for site {var}"
        )

    samples_dict = im.sample(
        num_samples=num_samples,
        rng_key=random.key(42),
        return_sites=["w", "b", "sigma"],
        return_datatree=False,
        X=X,
        y=y,
    )

    for k, v in samples_dict.items():
        assert v.shape[0] == num_samples, f"Incorrect number of samples for site {k}"


@pytest.mark.parametrize("mcmc", [lm], indirect=True)
def test_sample_with_mcmc(
    synthetic_data: tuple[ArrayLike, ArrayLike],
    mcmc: MCMC,
) -> None:
    """Test the `.sample()` method of ImpactModel with MCMC."""
    X, y = synthetic_data
    im = ImpactModel(lm, rng_key=random.key(42), inference=mcmc)
    im.fit_on_batch(X=X, y=y)

    num_samples = 7
    # rng_key is ignored for MCMC; sampling uses the post_warmup_state
    samples = im.sample(
        num_samples=num_samples,
        rng_key=random.key(42),
        X=X,
        y=y,
    ).posterior

    assert im.inference.num_samples == num_samples

    # Check shapes for all sampled sites
    for var in samples.data_vars:
        assert samples[var].values.shape[1] == num_samples, (
            f"Incorrect number of samples for site {var}"
        )

    samples_dict = im.sample(
        num_samples=num_samples,
        rng_key=random.key(42),
        return_datatree=False,
        X=X,
        y=y,
    )

    for k, v in samples_dict.items():
        assert v.shape[0] == num_samples, f"Incorrect number of samples for site {k}"



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MLflow model loading for aimz."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from mlflow.tracking.artifact_utils import _download_artifact_from_uri
from mlflow.utils.model_utils import (
    _add_code_from_conf_to_system_path,
    _get_flavor_configuration,
)

from aimz.mlflow.pyfunc import _load_model

if TYPE_CHECKING:
    from aimz.model.impact_model import ImpactModel

FLAVOR_NAME = "aimz"


def load_model(model_uri: str, dst_path: str | None = None) -> ImpactModel:
    """Load an aimz model from a local file or a run.

    Args:
        model_uri: The location, in URI format, of the MLflow model. For example:

            - ``/Users/me/path/to/local/model``
            - ``relative/path/to/local/model``
            - ``s3://my_bucket/path/to/model``
            - ``runs:/<mlflow_run_id>/run-relative/path/to/model``

            For more information about supported URI schemes, see
            `Referencing Artifacts <https://www.mlflow.org/docs/latest/tracking.html#
            artifact-locations>`_.

        dst_path: The local filesystem path to which to download the model artifact.
            This directory must already exist. If unspecified, a local output
            path will be created.

    Returns:
        An aimz model (an instance of :class:`~aimz.ImpactModel`).

    """
    local_model_path = _download_artifact_from_uri(model_uri, output_path=dst_path)
    flavor_conf = _get_flavor_configuration(local_model_path, FLAVOR_NAME)
    _add_code_from_conf_to_system_path(local_model_path, flavor_conf)
    model_artifacts_path = Path(local_model_path) / flavor_conf["pickled_model"]

    return _load_model(model_artifacts_path)



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for formatting and handling model outputs."""

import datetime
from importlib.metadata import version

import numpy as np
import xarray as xr
from jax import Array


def _make_attrs() -> dict[str, str]:
    """Generate metadata attributes for the aimz library.

    Returns:
        Attributes including creation timestamp and library version.
    """
    return {
        "created_at": datetime.datetime.now(datetime.UTC).isoformat(),
        "aimz_version": version("aimz"),
    }


def _dict_to_datatree(data: dict[str, Array]) -> xr.DataTree:
    """Convert a dictionary of arrays to an xarray DataTree.

    Each key in the dictionary becomes a variable in the Dataset, and its associated
    array is wrapped as an xarray DataArray with a ``chain`` and ``draw`` dimension to
    support MCMC-style outputs. Additional dimensions are automatically named using the
    pattern ``<variable>_dim_<N>``.

    Args:
        data: A dictionary mapping variable names to arrays. Each array should have
            shape ``(num_samples, dim_0, dim_1, ...)`` where the first dimension
            represents samples or draws.

    Returns:
        All variables with added ``chain`` and ``draw`` dimensions, along with
            coordinates for each array dimension.
    """
    return xr.DataTree(
        xr.Dataset(
            {
                site: xr.DataArray(
                    np.expand_dims(arr, axis=0),
                    coords={
                        "chain": np.arange(1),
                        "draw": np.arange(arr.shape[0]),
                        **{
                            f"{site}_dim_{i}": np.arange(arr.shape[i + 1])
                            for i in range(arr.ndim - 1)
                        },
                    },
                    dims=(
                        # Adding the 'chain' dimension to support MCMC-style structures.
                        "chain",
                        "draw",
                        # arr has shape (draw, dim_0, dim_1, ...), so arr.ndim includes
                        # 'draw' and we subtract 1
                        *[f"{site}_dim_{i}" for i in range(arr.ndim - 1)],
                    ),
                    name=site,
                )
                for site, arr in data.items()
            },
        ).assign_attrs(_make_attrs()),
    )



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MLflow logging for aimz models."""

from collections.abc import Iterable
from typing import TYPE_CHECKING

from mlflow.models import Model, ModelInputExample, ModelSignature
from mlflow.models.model import ModelInfo
from mlflow.tracking._model_registry import DEFAULT_AWAIT_MAX_SLEEP_SECONDS
from mlflow.utils.docstring_utils import LOG_MODEL_PARAM_DOCS, format_docstring

if TYPE_CHECKING:
    from aimz.model.impact_model import ImpactModel

FLAVOR_NAME = "aimz"


@format_docstring(LOG_MODEL_PARAM_DOCS.format(package_name=FLAVOR_NAME))
def log_model(
    model: "ImpactModel",
    conda_env: dict | None = None,
    code_paths: list | None = None,
    registered_model_name: str | None = None,
    signature: ModelSignature | None = None,
    input_example: ModelInputExample = None,
    await_registration_for: int | None = DEFAULT_AWAIT_MAX_SLEEP_SECONDS,
    pip_requirements: Iterable[str] | str | None = None,
    extra_pip_requirements: Iterable[str] | str | None = None,
    metadata: dict | None = None,
    name: str | None = None,
    params: dict[str, object] | None = None,
    model_type: str | None = None,
    model_id: str | None = None,
    step: int = 0,
    tags: dict[str, object] | None = None,
) -> ModelInfo:
    """Log an aimz model as an MLflow artifact for the current run..

    Args:
        model: An aimz model (an instance of :class:`~aimz.ImpactModel`) to be saved.
        conda_env: {{ conda_env }}
        code_paths: {{ code_paths }}
        registered_model_name: If given, each time a model is trained, it is registered
            as a new model version of the registered model with this name. The
            registered model is created if it does not already exist.
        signature: {{ signature }}
        input_example: {{ input_example }}
        await_registration_for: Number of seconds to wait for the model version to
            finish being created and is in ``READY`` status. By default, the function
            waits for five minutes. Specify ``0`` or :obj:`None` to skip waiting.
        pip_requirements: {{ pip_requirements }}
        extra_pip_requirements: {{ extra_pip_requirements }}
        metadata: {{ metadata }}
        name: {{ name }}
        params: {{ params }}
        model_type: {{ model_type }}
        model_id: {{ model_id }}
        step: {{ step }}
        tags: {{ tags }}

    Returns:
        A :external:class:`~mlflow.models.model.ModelInfo` instance that contains the
        metadata of the logged model.
    """
    import aimz.mlflow

    return Model.log(
        artifact_path=None,
        flavor=aimz.mlflow,
        registered_model_name=registered_model_name,
        await_registration_for=await_registration_for,
        metadata=metadata,
        model=model,
        name=name,
        model_type=model_type,
        params=params,
        tags=tags,
        step=step,
        model_id=model_id,
        conda_env=conda_env,
        code_paths=code_paths,
        signature=signature,
        input_example=input_example,
        pip_requirements=pip_requirements,
        extra_pip_requirements=extra_pip_requirements,
    )



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataclass for kernel metadata."""

from dataclasses import dataclass


@dataclass(frozen=True)
class KernelSpec:
    """A dataclass describing the kernel structure.

    Stores only the minimal information needed for downstream operations without
    re-tracing the NumPyro model.

    Attributes:
        traced:
            ``True`` once at least one successful trace has been performed.
        sample_sites:
            Names of latent (stochastic) sample sites encountered during tracing.
        return_sites:
            Names of default return sites. Always includes the output first, followed by
            any deterministic sites. Latent sample sites are excluded.
        output_observed:
            ``True`` if the output site was observed in the validating trace, used to
            distinguish prior-only from observed traces.

    Notes:
        Re-tracing only upgrades a prior-only spec (``output_observed`` is ``False``) to
        one that includes an observed output; the user kernel is assumed immutable after
        construction.
    """

    traced: bool
    sample_sites: tuple[str, ...]
    return_sites: tuple[str, ...]
    output_observed: bool



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for impact model."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Self

from aimz.utils._validation import _validate_kernel_signature

if TYPE_CHECKING:
    from collections.abc import Callable


class BaseModel(ABC):
    """Abstract base class for the impact model.

    Attributes:
        kernel (Callable): A probabilistic model with NumPyro primitives.
    """

    def __init__(
        self,
        kernel: Callable,
        param_input: str = "X",
        param_output: str = "y",
    ) -> None:
        """Initialize the BaseModel with a callable model.

        Args:
            kernel: A probabilistic model with NumPyro primitives.
            param_input: Name of the parameter in the ``kernel`` for the main input
                data.
            param_output: Name of the parameter in the ``kernel`` for the output data.
        """
        self._kernel = kernel
        self._param_input = param_input
        self._param_output = param_output
        _validate_kernel_signature(self._kernel, self._param_input, self._param_output)

    @abstractmethod
    def fit(self, X, y, **kwargs) -> Self:
        """Fit the model to the input data ``X`` and output data ``y``."""
        return self

    @abstractmethod
    def predict(self, X, **kwargs) -> object:
        """Predict the output based on the fitted model."""



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the `.estimate_effect()` method."""

from pathlib import Path

import jax.numpy as jnp
import pytest
from conftest import latent_variable_model, lm
from jax import random
from jax.typing import ArrayLike
from numpyro.infer import SVI, Trace_ELBO
from numpyro.infer.autoguide import AutoNormal
from numpyro.optim import Adam

from aimz import ImpactModel
from aimz._exceptions import NotFittedError


def test_model_not_fitted() -> None:
    """Calling `.estimate_effect()` on an unfitted model raises an error."""

    def kernel(X: ArrayLike, y: ArrayLike | None = None) -> None:
        pass

    im = ImpactModel(
        kernel,
        rng_key=random.key(42),
        inference=SVI(
            kernel,
            guide=AutoNormal(kernel),
            optim=Adam(step_size=1e-3),
            loss=Trace_ELBO(),
        ),
    )
    with pytest.raises(NotFittedError):
        im.estimate_effect()


@pytest.mark.parametrize("vi", [latent_variable_model], indirect=True)
def test_estimate_effect_argument_validation(
    synthetic_data: tuple[ArrayLike, ArrayLike],
    vi: SVI,
) -> None:
    """Validate argument exclusivity and successful effect computation."""
    X, y = synthetic_data
    im = ImpactModel(latent_variable_model, rng_key=random.key(42), inference=vi)
    im.fit(X=X, y=y, batch_size=len(X))

    msg = "Either `output_baseline` or `args_baseline` must be provided."
    with pytest.raises(ValueError, match=msg):
        im.estimate_effect(output_baseline=None, args_baseline=None)

    dt_baseline = im.predict_on_batch(X)

    msg = "Either `output_intervention` or `args_intervention` must be provided."
    with pytest.raises(ValueError, match=msg):
        im.estimate_effect(output_baseline=dt_baseline)

    dt_intervention = im.predict_on_batch(X, intervention={"z": jnp.zeros_like(y)})

    impact = im.estimate_effect(
        output_baseline=dt_baseline,
        output_intervention=dt_intervention,
    )

    assert impact.posterior_predictive["y"].mean(dim=["chain", "draw"]).shape == (
        len(y),
    )


@pytest.mark.parametrize("vi", [lm], indirect=True)
def test_estimate_effect_output_dir_lazy_args(
    synthetic_data: tuple[ArrayLike, ArrayLike],
    vi: SVI,
) -> None:
    """Ensure lazy (args_*) inputs work and baseline `output_dir` is propagated."""
    X, y = synthetic_data
    im = ImpactModel(lm, rng_key=random.key(42), inference=vi)
    im.fit(X=X, y=y, batch_size=len(X))

    msg = (
        r"The `batch_size` \(\d+\) is not divisible by the number of devices"
        r" \(\d+\)\."
    )
    with pytest.warns(UserWarning, match=msg):
        impact = im.estimate_effect(
            args_baseline={
                "X": X,
                "batch_size": len(X),
            },
            args_intervention={
                "X": X,
                "intervention": {"sigma": 10.0},
                "batch_size": len(X),
            },
        )

    # Confirm the temporary output directory propagated to effect result
    assert impact.attrs.get("output_dir") == str(Path(im.temp_dir).resolve())
    im.cleanup()



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MLflow integration for aimz."""

from aimz.mlflow.autolog import autolog
from aimz.mlflow.load import load_model
from aimz.mlflow.log import log_model
from aimz.mlflow.save import (
    get_default_conda_env,
    get_default_pip_requirements,
    save_model,
)

__all__ = [
    "autolog",
    "get_default_conda_env",
    "get_default_pip_requirements",
    "load_model",
    "log_model",
    "save_model",
]



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for custom data loader with padding logic for JAX arrays.

This module defines a custom ArrayLoader that processes batches of data and applies
padding to ensure the batch size is compatible with sharding across multiple XLA
devices.
"""

from __future__ import annotations

from math import ceil
from typing import TYPE_CHECKING
from warnings import warn

import jax.numpy as jnp
from jax import Array, device_put, local_device_count, random

if TYPE_CHECKING:
    from collections.abc import Iterator

    import numpy.typing as npt
    from jax.sharding import Sharding
    from jax.typing import ArrayLike

    from aimz.utils.data.array_dataset import ArrayDataset


class ArrayLoader:
    """ArrayLoader class for JAX arrays."""

    def __init__(
        self,
        dataset: ArrayDataset,
        rng_key: ArrayLike,
        *,
        batch_size: int = 32,
        shuffle: bool = False,
        device: Sharding | None = None,
    ) -> None:
        """Initialize an ArrayLoader instance.

        Args:
            dataset: The dataset to load.
            rng_key (ArrayLike): A pseudo-random number generator key.
            batch_size: The number of samples per batch.
            shuffle: Whether to shuffle the dataset before batching.
            device: The device or sharding specification to which the data should be
                moved. By default, no device transfer is applied. When used as an input
                to a model, this will be overridden by the device setting of the model.
        """
        self.dataset = dataset
        if (
            not isinstance(batch_size, int)
            or isinstance(batch_size, bool)
            or batch_size <= 0
        ):
            msg = f"`batch_size` should be a positive integer, but got {batch_size!r}."
            raise ValueError(msg)
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.indices = jnp.arange(len(self.dataset))
        if isinstance(rng_key, Array) and rng_key.dtype == jnp.uint32:
            msg = "Legacy `uint32` PRNGKey detected; converting to a typed key array."
            warn(msg, category=UserWarning, stacklevel=2)
            rng_key = random.wrap_key_data(rng_key)
        self.rng_key = rng_key
        self._num_devices = local_device_count()
        self.device = device

    def pad_array(self, x: Array | npt.NDArray, n_pad: int, axis: int = 0) -> Array:
        """Pad an array to ensure compatibility with sharding.

        Args:
            x: The input array to be padded.
            n_pad: The number of padding elements to add.
            axis: The axis along which to apply the padding.

        Returns:
            The padded array with padding applied along the specified axis.

        Raises:
            ValueError: If padding is requested along an unsupported axis for a 1D
                array.
        """
        if x.ndim == 1:
            if axis == 0:
                return jnp.pad(x, pad_width=(0, n_pad), mode="edge")
            msg = "Padding 1D arrays is only supported along axis 0."
            raise ValueError(msg)

        # Initialize all axes with no padding
        pad_width: list[tuple[int, int]] = [(0, 0)] * x.ndim
        # Apply padding to the specified axis
        pad_width[axis] = (0, n_pad)

        return jnp.pad(x, pad_width=pad_width, mode="edge")

    def __iter__(self) -> Iterator[tuple[dict[str, Array], int]]:
        """Iterate over the dataset in batches.

        Yields:
            A batch of arrays with data from the dataset.
            The number of padded samples added for sharding compatibility.
        """
        indices = self.indices
        if self.shuffle:
            self.rng_key, subkey = random.split(self.rng_key)
            indices = random.permutation(subkey, self.indices)
        for start in range(0, len(self.dataset), self.batch_size):
            end = start + self.batch_size
            batch_idx = indices[start:end]
            if self.device is not None:
                n_pad = (-len(batch_idx)) % self._num_devices
                batch = {
                    k: self.pad_array(arr[batch_idx], n_pad=n_pad)
                    for k, arr in self.dataset.arrays.items()
                }
                batch = {k: device_put(v, self.device) for k, v in batch.items()}
            else:
                n_pad = 0
                batch = {k: arr[batch_idx] for k, arr in self.dataset.arrays.items()}
            yield batch, n_pad

    def __len__(self) -> int:
        """Return the number of batches.

        Returns:
            The total number of batches.
        """
        return ceil(len(self.dataset) / self.batch_size)



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for validating models and objects."""

from __future__ import annotations

from inspect import Parameter, getfullargspec, signature
from typing import TYPE_CHECKING

import jax.numpy as jnp
from jax import Array
from sklearn.utils.validation import check_array, check_X_y

from aimz._exceptions import KernelValidationError, NotFittedError

if TYPE_CHECKING:
    from collections import OrderedDict
    from collections.abc import Callable

    import xarray as xr
    from jax.typing import ArrayLike

    from aimz import ImpactModel


def _check_is_fitted(model: ImpactModel, msg: str | None = None) -> None:
    """Check if the model is fitted.

    Raises:
        NotFittedError: If the model has not been fitted.
    """
    if msg is None:
        msg = (
            "This %(name)s instance is not fitted yet. Call ``.fit()`` with "
            "appropriate arguments before using the model."
        )
    if not _is_fitted(model):
        raise NotFittedError(msg % {"name": type(model).__name__})


def _is_fitted(model: ImpactModel) -> bool:
    if hasattr(model, "_is_fitted"):
        return model.is_fitted()

    return any(v.endswith("_") and not v.startswith("__") for v in vars(model))


def _validate_group(dt_baseline: xr.DataTree, dt_intervention: xr.DataTree) -> str:
    """Validate the groups in ``dt_baseline`` and ``dt_intervention``.

    Args:
        dt_baseline: Precomputed output for the baseline scenario.
        dt_intervention: Precomputed output for the intervention scenario.

    Returns:
        The group name (``predictions`` or ``posterior_predictive``).

    Raises:
        ValueError: If the group is not found in ``dt_intervention``.
    """
    group = (
        "predictions"
        if "predictions" in dt_baseline.children
        else "posterior_predictive"
    )

    if group not in dt_intervention.children:
        sub = group
        msg = (
            f"Group {sub!r} not found in `dt_intervention`. Available "
            f"groups: {', '.join(map(repr, dt_intervention.children))}"
        )
        raise ValueError(msg)

    return group


def _validate_kernel_signature(
    kernel: Callable,
    param_input: str,
    param_output: str,
) -> None:
    """Validate the signature of a kernel function.

    Args:
        kernel: The kernel function to validate.
        param_input: Name of the parameter in ``kernel`` corresponding to the input.
        param_output: Name of the parameter in ``kernel`` corresponding to the output.

    Raises:
        KernelValidationError: If the kernel signature does not meet the required
            constraints.
    """
    argspec = getfullargspec(kernel)
    if argspec.varargs is not None or argspec.varkw is not None:
        msg = "Kernel must not accept variable arguments (*args or **kwargs)."
        raise KernelValidationError(msg)

    param_main = [
        arg
        for arg in (param_input, param_output)
        if arg not in (argspec.args + argspec.kwonlyargs)
    ]
    if param_main:
        sub = ", ".join(map(repr, param_main))
        msg = (
            f"Kernel must accept {sub} as argument(s). Modify the kernel signature or "
            "set `param_input` and param_output` accordingly."
        )
        raise KernelValidationError(msg)

    sig = signature(kernel)
    if sig.parameters[param_input].default is not Parameter.empty:
        sub = param_input
        msg = f"{sub!r} must not have a default value."
        raise KernelValidationError(msg)
    if sig.parameters[param_output].default:
        sub = param_output
        msg = f"{sub!r} must have a default value of `None`."
        raise KernelValidationError(msg)


def _validate_kernel_body(
    kernel: Callable,
    *,
    param_output: str,
    model_trace: OrderedDict[str, dict],
    with_output: bool,
) -> None:
    """Validate the body of a kernel function.

    Args:
        kernel: The kernel function to validate.
        param_output: Name of the parameter in ``kernel`` corresponding to the output.
        model_trace: The model trace containing the sites.
        with_output: Whether the kernel is expected to have observed output.

    Raises:
        KernelValidationError: If the kernel body does not meet the required
            constraints.
    """
    invalid_site = [site for site in model_trace if "/" in site]
    if invalid_site:
        msg = (
            f"Invalid site names containing '/': {invalid_site!r}. "
            "xarray.DataTree does not allow '/' in variable names."
        )
        raise KernelValidationError(msg)

    if param_output not in model_trace:
        msg = f"Kernel must include a sample site named {param_output!r}."
        raise KernelValidationError(msg)
    site = model_trace[param_output]
    if site["type"] != "sample":
        msg = f"Expected {param_output!r} to have type 'sample', got {site['type']!r}."
        raise KernelValidationError(msg)
    if with_output and not site.get("is_observed", False):
        msg = (
            f"{param_output!r} must be observed (i.e., defined with `obs=` in the "
            "kernel)."
        )
        raise KernelValidationError(msg)

    # Collect parameter names from the kernel signature, excluding the output parameter
    params = getfullargspec(kernel).args + getfullargspec(kernel).kwonlyargs
    params.remove(param_output)
    # Check for name conflicts between parameter names and model site names
    conflicts = set(params) & set(model_trace.keys())
    if conflicts:
        msg = (
            f"Kernel parameters conflict with model sites: "
            f"{', '.join(repr(k) for k in sorted(conflicts))}. "
            "Rename parameters or revise the model to avoid shadowing."
        )
        raise KernelValidationError(msg)


def _validate_X_y_to_jax(
    X: ArrayLike,
    y: ArrayLike | None = None,
) -> tuple[Array, Array] | Array:
    """Validate and convert data arrays to JAX arrays.

    Arrays are checked, converted, and placed on the same device as their originals
    when available.

    Args:
        X (ArrayLike): Input data with shape ``(n_samples_X, n_features)``.
        y (ArrayLike): Output data with shape ``(n_samples_Y,)``.

    Returns:
        Validated JAX arrays, returning ``X`` if only X is provided, or a tuple
        ``(X, y)`` otherwise.
    """
    if y is None:
        device_x = X.device if isinstance(X, Array) and X.committed else None

        return jnp.asarray(check_array(X), device=device_x)

    device_x, device_y = (
        arr.device if isinstance(arr, Array) and arr.committed else None
        for arr in (X, y)
    )
    X, y = check_X_y(X, y, force_writeable=True, y_numeric=True)

    return jnp.asarray(X, device=device_x), jnp.asarray(y, device=device_y)



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MLflow autologging integration for aimz."""

import logging
import tempfile
from collections.abc import Callable
from inspect import getsource
from typing import TYPE_CHECKING, cast

import numpy as np
from mlflow import log_metric, log_param, log_params, log_text
from mlflow.models import infer_signature
from mlflow.tracking.fluent import _initialize_logged_model
from mlflow.utils.autologging_utils import (
    INPUT_EXAMPLE_SAMPLE_ROWS,
    autologging_integration,
    get_autologging_config,
    log_fn_args_as_params,
    resolve_input_example_and_signature,
    safe_patch,
)
from sklearn.utils.validation import _is_arraylike

if TYPE_CHECKING:
    from numpyro.infer.svi import SVIRunResult

FLAVOR_NAME = "aimz"

_logger = logging.getLogger(__name__)


@autologging_integration(FLAVOR_NAME)
def autolog(
    *,
    log_input_examples: bool = False,
    log_model_signatures: bool = True,
    log_models: bool = True,
    disable: bool = False,
    silent: bool = False,
    registered_model_name: str | None = None,
    extra_tags: dict[str, str] | None = None,
) -> None:
    """Enable and configure autologging for aimz with MLflow.

    Autologging is performed when you call:
        - :meth:`~aimz.ImpactModel.fit_on_batch`.
        - :meth:`~aimz.ImpactModel.fit`.

    Logs the following:
        - Selected arguments to the methods, together with ``param_input``,
          ``param_output``, ``inference_method``, and ``optimizer`` as parameters.
        - The final evidence lower bound (ELBO) loss as a metric.
        - The source code of the kernel function used in the model as a text artifact.
        - An MLflow Model containing the fitted estimator as an artifact.

    Args:
        log_input_examples: If ``True``, input examples from training datasets are
            collected and logged along with model artifacts during training. If
            ``False``, input examples are not logged. Note: Input examples are MLflow
            model attributes and are only collected if ``log_models`` is also ``True``.
        log_model_signatures: If ``True``,
            :external:class:`ModelSignatures <mlflow.models.ModelSignature>` describing
            model inputs and outputs are collected and logged along with model artifacts
            during training. If ``False``, signatures are not logged. Note: Model
            signatures are MLflow model attributes and are only collected if
            ``log_models`` is also ``True``.
        log_models: If ``True``, trained models are logged as MLflow model artifacts. If
            ``False``, trained models are not logged. Input examples and model
            signatures, which are attributes of MLflow models, are also omitted when
            ``log_models`` is ``False``.
        disable: If ``True``, disables all supported autologging integrations. If
            ``False``, enables all supported autologging integrations.
        silent: If ``True``, suppress all event logs and warnings from MLflow during
            autologging setup and training execution. If ``False``, show all events and
            warnings during autologging setup and training execution.
        registered_model_name: If given, each time a model is trained, it is registered
            as a new model version of the registered model with this name. The
            registered model is created if it does not already exist.
        extra_tags: A dictionary of extra tags to set on each managed run created by
            autologging.
    """
    from aimz.mlflow.log import log_model
    from aimz.model.impact_model import ImpactModel
    from aimz.utils.data import ArrayLoader

    def patch_fit(
        original: Callable,
        self: ImpactModel,
        *args: object,
        **kwargs: object,
    ) -> ImpactModel:
        """Patch for the fitting method to log information.

        Args:
            original (Callable): The original method.
            self (ImpactModel): The instance being initialized.
            *args (object): Positional arguments for the method.
            **kwargs (object): Keyword arguments for the method.
        """
        model_source = getsource(self.kernel)
        log_text(model_source, artifact_file="model.py")

        unlogged = ["X", "y", "num_samples", "progress", "kwargs"]
        params = {
            "param_input": self.param_input,
            "param_output": self.param_output,
            "inference_method": type(self.inference).__name__,
        }
        if params["inference_method"] == "SVI":
            params["optimizer"] = type(self.inference.optim).__name__
        log_params(params)
        log_fn_args_as_params(
            original,
            args=args,
            kwargs={k: v for k, v in kwargs.items() if not _is_arraylike(v)},
            unlogged=unlogged,
        )

        model = original(self, *args, **kwargs)

        log_param("num_samples", self._num_samples)
        if params["inference_method"] == "SVI":
            losses = cast("SVIRunResult", self.vi_result).losses
            log_metric("elbo_loss", value=losses[-1])

        if log_models:
            model_id = _initialize_logged_model("model", flavor=FLAVOR_NAME).model_id
            # Will only resolve `input_example` and `signature` if `log_models` is
            # `True`.
            with tempfile.TemporaryDirectory() as temp_dir:
                X = kwargs.get("X") if "X" in kwargs else args[0]
                if isinstance(X, ArrayLoader):
                    input_example = {
                        k: np.asarray(v[:INPUT_EXAMPLE_SAMPLE_ROWS])
                        for k, v in X.dataset.arrays.items()
                        if k != self.param_output
                    }
                else:
                    input_example = {
                        "X": np.asarray(X)[:INPUT_EXAMPLE_SAMPLE_ROWS],
                        **{
                            k: np.asarray(v)[:INPUT_EXAMPLE_SAMPLE_ROWS]
                            for k, v in kwargs.items()
                            if k != self.param_output and _is_arraylike(v)
                        },
                    }
                    if len(input_example) == 1:
                        input_example = input_example["X"]
                input_example, signature = resolve_input_example_and_signature(
                    get_input_example=lambda: input_example,
                    infer_model_signature=lambda x: infer_signature(
                        x,
                        params={
                            "batch_size": INPUT_EXAMPLE_SAMPLE_ROWS,
                            "output_dir": temp_dir,
                            "progress": False,
                        },
                    ),
                    log_input_example=log_input_examples,
                    log_model_signature=log_model_signatures,
                    logger=_logger,
                )
            registered_model_name = get_autologging_config(
                FLAVOR_NAME,
                config_key="registered_model_name",
                default_value=None,
            )
            log_model(
                model,
                registered_model_name=registered_model_name,
                signature=signature,
                input_example=input_example,
                model_id=model_id,
            )
            if params["inference_method"] == "SVI":
                log_metric("elbo_loss", value=losses[-1], model_id=model_id)

        return model

    safe_patch(
        FLAVOR_NAME,
        destination=ImpactModel,
        function_name="fit_on_batch",
        patch_function=patch_fit,
        manage_run=True,
        extra_tags=extra_tags,
    )

    safe_patch(
        FLAVOR_NAME,
        destination=ImpactModel,
        function_name="fit",
        patch_function=patch_fit,
        manage_run=True,
        extra_tags=extra_tags,
    )



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MLflow pyfunc wrapper for aimz models."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, cast

from mlflow.pyfunc import PyFuncModel

if TYPE_CHECKING:
    from jax.typing import ArrayLike

    from aimz.model.impact_model import ImpactModel
    from aimz.utils.data import ArrayLoader


class _ImpactModelWrapper(PyFuncModel):
    """MLflow PyFunc wrapper for ImpactModel."""

    def __init__(self, model: ImpactModel) -> None:
        self.model = model

    def get_raw_model(self) -> ImpactModel:
        """Return the underlying ImpactModel instance."""
        return self.model

    def predict(self, model_input: object, params: dict | None = None):
        """Run predictions using the wrapped ImpactModel."""
        if isinstance(model_input, dict):
            return self.model.predict(**model_input, **(params or {}))
        return self.model.predict(
            cast("ArrayLike | ArrayLoader", model_input),
            **(params or {}),
        )


def _load_pyfunc(path: str) -> _ImpactModelWrapper:
    """Load the ImpactModel as an MLflow PyFunc wrapper."""
    return _ImpactModelWrapper(_load_model(path))


def _load_model(path: str) -> ImpactModel:
    """Load an ImpactModel instance from the specified path."""
    import cloudpickle

    with Path(path).open("rb") as f:
        return cloudpickle.load(f)



# Copyright 2025 Eli Lilly and Company
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for initializing inputs and preprocessing arguments for data pipelines."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, NamedTuple, cast
from warnings import warn

from jax.typing import ArrayLike
from sklearn.utils.validation import check_array, check_X_y

from aimz.utils._kwargs import _group_kwargs
from aimz.utils.data import ArrayDataset, ArrayLoader

if TYPE_CHECKING:
    from collections.abc import Sized

    from jax.sharding import Sharding

logger = logging.getLogger(__name__)

# Maximum number of elements allowed per batch or chunk. Each element is assumed to be
# 4 bytes (float32 or int32). `MAX_ELEMENTS = 25_000_000` corresponds to ~100 MB per
# chunk, helping to control disk and memory usage during data processing.
MAX_ELEMENTS = 25_000_000


def _setup_inputs(
    *,
    X: ArrayLike | ArrayLoader,
    y: ArrayLike | None,
    rng_key: ArrayLike,
    batch_size: int | None,
    num_samples: int,
    shuffle: bool = False,
    device: Sharding | None = None,
    **kwargs: object,
) -> tuple[ArrayLoader, NamedTuple]:
    """Prepare an dataloader and grouped keyword arguments.

    Args:
        X (ArrayLike | ArrayLoader): Input data, either an array-like of shape
            ``(n_samples, n_features)`` or a data loader that holds all array-like
            objects and handles batching internally; if a data loader is passed,
            ``rng_key``, ``batch_size`` and ``shuffle`` are ignored.
        y (ArrayLike | None): Output data with shape ``(n_samples_Y,)``. Must be
            ``None`` if ``X`` is a data loader.
        rng_key (ArrayLike): A pseudo-random number generator key.
        batch_size: The size of batches for data loading.
        num_samples: Number of samples to draw, which affects the size of batches.
        shuffle: Whether to shuffle the dataset before batching.
        device: The device or sharding specification to which the data should be moved.
            By default, no device transfer is applied. If ``X`` is a data loader, it
            will override the device setting of the loader.
        **kwargs: Additional arguments passed to the model. All array-like values are
            expected to be JAX arrays.

    Returns:
        - The data loader for batching.
        - Extra keyword arguments to be passed downstream.
    """
    kwargs_array, kwargs_extra = _group_kwargs(kwargs)

    if isinstance(X, ArrayLike):
        if y is None:
            X = check_array(X)
        else:
            X, y = check_X_y(X, y, force_writeable=True, y_numeric=True)
        num_devices = device.num_devices if device else 1
        if batch_size is None:
            if len(cast("Sized", X)) * num_samples < MAX_ELEMENTS:
                batch_size = len(cast("Sized", X))
            else:
                batch_size = MAX_ELEMENTS // num_samples
                batch_size -= batch_size % num_devices
            logger.info(
                "Using batch_size=%d. Specify explicitly to better control memory "
                "usage.",
                batch_size,
            )
        if batch_size % num_devices != 0:
            msg = (
                f"The `batch_size` ({batch_size}) is not divisible by the number of "
                f"devices ({num_devices}). Use a multiple of {num_devices} "
                "for optimal performance."
            )
            warn(msg, category=UserWarning, stacklevel=2)
        loader = ArrayLoader(
            ArrayDataset(X=cast("ArrayLike", X), y=y, **kwargs_array._asdict()),
            rng_key=rng_key,
            batch_size=batch_size,
            shuffle=shuffle,
            device=device,
        )
    elif isinstance(X, ArrayLoader):
        if y is not None:
            msg = "`y` must be `None` when `X` is already a data loader."
            raise TypeError(msg)
        loader = X
        loader.device = device

    return loader, kwargs_extra
